@@ -1,10 +1,11 @@
+# -*- coding: utf-8 -*-
+
 """Populate an item's `added` and `mtime` fields by using the file
 modification time (mtime) of the item's source file before import.
 
 Reimported albums and items are skipped.
 """
-from __future__ import (unicode_literals, absolute_import, print_function,
-                        division)
+from __future__ import division, absolute_import, print_function
 
 import os
 
@@ -18,6 +19,7 @@ class ImportAddedPlugin(BeetsPlugin):
         super(ImportAddedPlugin, self).__init__()
         self.config.add({
             'preserve_mtimes': False,
+            'preserve_write_mtimes': False,
         })
 
         # item.id for new items that were reimported
@@ -34,8 +36,10 @@ class ImportAddedPlugin(BeetsPlugin):
         register('before_item_moved', self.record_import_mtime)
         register('item_copied', self.record_import_mtime)
         register('item_linked', self.record_import_mtime)
+        register('item_hardlinked', self.record_import_mtime)
         register('album_imported', self.update_album_times)
         register('item_imported', self.update_item_times)
+        register('after_write', self.update_after_write_time)
 
     def check_config(self, task, session):
         self.config['preserve_mtimes'].get(bool)
@@ -48,7 +52,7 @@ class ImportAddedPlugin(BeetsPlugin):
 
     def record_if_inplace(self, task, session):
         if not (session.config['copy'] or session.config['move'] or
-                session.config['link']):
+                session.config['link'] or session.config['hardlink']):
             self._log.debug(u"In place import detected, recording mtimes from "
                             u"source paths")
             items = [task.item] \
@@ -59,7 +63,7 @@ class ImportAddedPlugin(BeetsPlugin):
 
     def record_reimported(self, task, session):
         self.reimported_item_ids = set(item.id for item, replaced_items
-                                       in task.replaced_items.iteritems()
+                                       in task.replaced_items.items()
                                        if replaced_items)
         self.replaced_album_paths = set(task.replaced_albums.keys())
 
@@ -73,11 +77,6 @@ class ImportAddedPlugin(BeetsPlugin):
         """Write the given mtime to an item's `mtime` field and to the mtime
         of the item's file.
         """
-        if mtime is None:
-            self._log.warn(u"No mtime to be preserved for item '{0}'",
-                           util.displayable_path(item.path))
-            return
-
         # The file's mtime on disk must be in sync with the item's mtime
         self.write_file_mtime(util.syspath(item.path), mtime)
         item.mtime = mtime
@@ -124,3 +123,13 @@ class ImportAddedPlugin(BeetsPlugin):
             self._log.debug(u"Import of item '{0}', selected item.added={1}",
                             util.displayable_path(item.path), item.added)
             item.store()
+
+    def update_after_write_time(self, item):
+        """Update the mtime of the item's file with the item.added value
+        after each write of the item if `preserve_write_mtimes` is enabled.
+        """
+        if item.added:
+            if self.config['preserve_write_mtimes'].get(bool):
+                self.write_item_mtime(item, item.added)
+            self._log.debug(u"Write of item '{0}', selected item.added={1}",
+                            util.displayable_path(item.path), item.added)
